from dataclasses import dataclass, field
from typing import Dict

from dnn_cool.visitors import LeafVisitor, VisitorOut, RootCompositeVisitor


class ActivationVisitor(LeafVisitor):

    def __init__(self, task, prefix):
        super().__init__(task, prefix)

    def full_result(self, preds, targets):
        return ActivationsFiltered({self.path: preds})

    def empty_result(self):
        return ActivationsFiltered({self.path: {}})

    def preconditioned_result(self, preds, targets):
        return ActivationsFiltered({self.path: preds})


@dataclass
class ActivationsFiltered(VisitorOut):
    data: Dict = field(default_factory=lambda: {})

    def __iadd__(self, other):
        self.data.update(other.data)
        return self

    def reduce(self):
        return self.data


class CompositeActivation(RootCompositeVisitor):

    def __init__(self, task_flow):
        super().__init__(task_flow, ActivationVisitor, ActivationsFiltered)



import os
from dataclasses import dataclass, field
from typing import Iterable, Optional, Callable, Tuple, Any, Sequence

import torch
from torch import nn
from torch.utils.data import Dataset

from dnn_cool.activations import CompositeActivation
from dnn_cool.datasets import FlowDataset, LeafTaskDataset
from dnn_cool.decoders import BinaryDecoder, TaskFlowDecoder, Decoder, ClassificationDecoder, \
    MultilabelClassificationDecoder
from dnn_cool.evaluation import EvaluationCompositeVisitor, EvaluationVisitor
from dnn_cool.filter import FilterCompositeVisitor, FilterVisitor
from dnn_cool.losses import TaskFlowLoss, ReducedPerSample, TaskFlowLossPerSample
from dnn_cool.metrics import TorchMetric, get_default_binary_metrics, \
    get_default_bounded_regression_metrics, get_default_classification_metrics, \
    get_default_multilabel_classification_metrics
from dnn_cool.missing_values import positive_values
from dnn_cool.modules import SigmoidAndMSELoss, Identity, TaskFlowModule
from dnn_cool.treelib import TreeExplainer


class ITask:

    def __init__(self, name: str, inputs, available_func=None, metrics: Tuple[str, TorchMetric] = ()):
        self.name = name
        self.inputs = inputs
        self.available_func = available_func
        self.metrics = metrics

    def get_name(self) -> str:
        return self.name

    def get_activation(self) -> Optional[nn.Module]:
        return None

    def get_decoder(self) -> Optional[Decoder]:
        return None

    def get_filter(self) -> FilterVisitor:
        return FilterVisitor(self, prefix='')

    def get_evaluator(self) -> EvaluationVisitor:
        return EvaluationVisitor(self, prefix='')

    def has_children(self) -> bool:
        return False

    def get_available_func(self):
        return self.available_func

    def get_loss(self):
        raise NotImplementedError()

    def get_per_sample_loss(self):
        raise NotImplementedError()

    def torch(self):
        raise NotImplementedError()

    def get_inputs(self):
        return self.inputs

    def get_labels(self):
        raise NotImplementedError()

    def get_dataset(self):
        raise NotImplementedError()

    def get_metrics(self):
        for i in range(len(self.metrics)):
            metric_name, metric = self.metrics[i]
            metric.bind_to_task(self)
        return self.metrics

    def __repr__(self):
        params = (
            ('name', self.get_name()),
            ('module', self.torch()),
            ('loss', self.get_loss()),
            ('activation', self.get_activation()),
            ('decoder', self.get_decoder()),
            ('available_func', self.get_available_func()),
            ('per_sample_loss', self.get_per_sample_loss())
        )
        params_str = os.linesep.join(map(lambda x: f'\t{x[0]}={x[1]}', params))
        params_str = f'{os.linesep}{params_str}{os.linesep}'
        res = f'{self.__class__.__module__}.{self.__class__.__name__}({params_str}) at {hex(id(self))}'
        return res


@dataclass
class Task(ITask):
    name: str
    labels: Sequence
    loss: nn.Module
    per_sample_loss: nn.Module
    available_func: Callable
    inputs: Sequence
    activation: Optional[nn.Module]
    decoder: Decoder
    module: nn.Module
    metrics: Tuple[str, TorchMetric]

    def get_activation(self) -> Optional[nn.Module]:
        return self.activation

    def get_decoder(self):
        return self.decoder

    def get_loss(self):
        return self.loss

    def get_per_sample_loss(self):
        return self.per_sample_loss

    def torch(self):
        return self.module

    def get_inputs(self, *args, **kwargs):
        return self.inputs

    def get_labels(self, *args, **kwargs):
        return self.labels

    def get_dataset(self, **kwargs):
        return LeafTaskDataset(self.inputs, self.labels)


@dataclass()
class BinaryHardcodedTask(Task):
    name: str
    labels: Sequence
    loss: nn.Module = None
    per_sample_loss: nn.Module = None
    available_func: Callable = positive_values
    inputs: Sequence = None
    activation: Optional[nn.Module] = None
    decoder: Decoder = None
    module: nn.Module = Identity()
    metrics: Tuple[str, TorchMetric] = ()


@dataclass()
class BoundedRegressionTask(Task):
    """
    Represents a regression task, where the labels are normalized between 0 and 1. Examples include bounding box top
    left corners regression. Here are the defaults:
    * activation - `nn.Sigmoid()` - so that the output is in `[0, 1]`
    * loss - `SigmoidAndMSELoss` - sigmoid on the logits, then standard mean squared error loss.
    """
    name: str
    labels: Sequence
    loss: nn.Module = field(default_factory=lambda: SigmoidAndMSELoss(reduction='mean'))
    per_sample_loss: nn.Module = field(default_factory=lambda: ReducedPerSample(SigmoidAndMSELoss(reduction='none'),
                                                                                reduction=torch.sum))
    available_func: Callable = field(default_factory=lambda: positive_values)
    module: nn.Module = field(default_factory=lambda: nn.Identity())
    activation: nn.Module = field(default_factory=lambda: nn.Sigmoid())
    decoder: Decoder = None
    inputs: Sequence = None
    metrics: Tuple = field(default_factory=get_default_bounded_regression_metrics)


@dataclass()
class BinaryClassificationTask(Task):
    """
    Represents a normal binary classification task. Labels should be between 0 and 1.
    * activation - `nn.Sigmoid()`
    * loss - ``nn.BCEWithLogitsLoss()`
    """
    name: str
    labels: Sequence
    loss: nn.Module = nn.BCEWithLogitsLoss(reduction='mean')
    per_sample_loss: nn.Module = ReducedPerSample(nn.BCEWithLogitsLoss(reduction='none'), reduction=torch.mean)
    available_func: Callable = positive_values
    inputs: Sequence = None
    activation: Optional[nn.Module] = nn.Sigmoid()
    decoder: Decoder = field(default_factory=BinaryDecoder)
    module: nn.Module = Identity()
    metrics: Tuple[str, TorchMetric] = field(default_factory=get_default_binary_metrics)


@dataclass
class ClassificationTask(Task):
    """
    Represents a classification task. Labels should be integers from 0 to N-1, where N is the number of classes
    * activation - `nn.Softmax(dim=-1)`
    * loss - `nn.CrossEntropyLoss()`
    """
    name: str
    labels: Sequence
    loss: nn.Module = nn.CrossEntropyLoss(reduction='mean')
    per_sample_loss: nn.Module = ReducedPerSample(nn.CrossEntropyLoss(reduction='none'), torch.mean)
    available_func: Callable = positive_values
    inputs: Sequence = None
    activation: Optional[nn.Module] = nn.Softmax()
    decoder: Decoder = field(default_factory=ClassificationDecoder)
    module: nn.Module = Identity()
    metrics: Tuple[str, TorchMetric] = field(default_factory=get_default_classification_metrics)


@dataclass()
class MultilabelClassificationTask(Task):
    """
    Represents a classification task. Labels should be integers from 0 to N-1, where N is the number of classes
    * activation - `nn.Softmax(dim=-1)`
    * loss - `nn.CrossEntropyLoss()`
    """
    name: str
    labels: Sequence
    loss: nn.Module = nn.BCEWithLogitsLoss(reduction='mean')
    per_sample_loss: nn.Module = ReducedPerSample(nn.BCEWithLogitsLoss(reduction='none'), torch.mean)
    available_func: Callable = positive_values
    inputs: Sequence = None
    activation: Optional[nn.Module] = nn.Sigmoid()
    decoder: Decoder = field(default_factory=MultilabelClassificationDecoder)
    module: nn.Module = Identity()
    metrics: Tuple[str, TorchMetric] = field(default_factory=get_default_multilabel_classification_metrics)


class TaskFlow(ITask):

    def __init__(self, name, tasks: Iterable[ITask], flow_func=None, inputs=None, available_func=None):
        super().__init__(name, inputs, available_func)
        self.tasks = {}
        for task in tasks:
            self.tasks[task.get_name()] = task
        if flow_func is not None:
            self._flow_func = flow_func

    def get_loss(self):
        return TaskFlowLoss(self)

    def get_per_sample_loss(self):
        return TaskFlowLossPerSample(self)

    def torch(self):
        return TaskFlowModule(self)

    def has_children(self):
        return True

    def get_dataset(self, **kwargs) -> Dataset:
        return FlowDataset(self, **kwargs)

    def flow(self, x, out):
        raise NotImplementedError()

    def get_flow_func(self):
        if hasattr(self, '_flow_func') and self._flow_func is not None:
            return self._flow_func
        return self.__class__.flow

    def get_metrics(self):
        all_metrics = []
        for task in self.tasks.values():
            all_metrics += task.get_metrics()
        return all_metrics

    def get_treelib_explainer(self):
        return TreeExplainer(self)

    def get_labels(self, *args, **kwargs):
        all_labels = []
        for task in self.tasks.values():
            all_labels += task.get_labels()
        return all_labels

    def get_decoder(self):
        return TaskFlowDecoder(self)

    def get_activation(self) -> Callable:
        return CompositeActivation(self)

    def get_evaluator(self):
        return EvaluationCompositeVisitor(self, prefix='')

    def get_filter(self):
        return FilterCompositeVisitor(self, prefix='')

    def get_all_children(self, prefix=''):
        tasks = {}
        for task_name, task in self.tasks.items():
            if task.has_children():
                assert isinstance(task, TaskFlow)
                tasks.update(task.get_all_children(prefix=f'{prefix}{task.get_name()}.'))
            else:
                tasks[prefix + task_name] = task
        return tasks



from dataclasses import dataclass, field
from typing import List

import pandas as pd
import torch

from dnn_cool.visitors import LeafVisitor, VisitorOut, RootCompositeVisitor


class EvaluationVisitor(LeafVisitor):

    def __init__(self, task, prefix):
        super().__init__(task, prefix)
        self.metrics = task.get_metrics()

    def full_result(self, preds, targets):
        return self.compute_metrics(preds, targets)

    def empty_result(self):
        return EvaluationResults([])

    def preconditioned_result(self, preds, targets):
        return self.compute_metrics(preds, targets)

    def compute_metrics(self, preds, targets):
        res = []
        for metric_name, metric in self.metrics:
            # No activation, since preds is already activated
            metric_res = metric(preds, targets, activate=False)
            if isinstance(metric_res, dict):
                for key, value in metric_res.items():
                    res.append(self.create_evaluation_record(f'{metric_name}_{key}', value, targets))
            else:
                res.append(self.create_evaluation_record(metric_name, metric_res, targets))
        return EvaluationResults(res)

    def create_evaluation_record(self, metric_name, metric_res, targets):
        if isinstance(metric_res, torch.Tensor):
            metric_res = metric_res.item()
        return {
            'task_path': self.path,
            'metric_name': metric_name,
            'metric_res': metric_res,
            'num_samples': len(targets),
        }


@dataclass
class EvaluationResults(VisitorOut):
    data: List = field(default_factory=lambda: [])

    def __iadd__(self, other):
        self.data += other.data
        return self

    def reduce(self):
        return pd.DataFrame(self.data)


class EvaluationCompositeVisitor(RootCompositeVisitor):

    def __init__(self, task_flow, prefix):
        super().__init__(task_flow, EvaluationVisitor, EvaluationResults, prefix=prefix)

    def load_tuned(self, tuned_params):
        tasks = self.task_flow.get_all_children()
        for path, task in tasks.items():
            task.get_decoder().load_tuned(tuned_params[path])



import torch
from torch import nn

from dnn_cool.dsl import IFeaturesDict, IOut, ICondition, IFlowTaskResult
from dnn_cool.utils import any_value


class ReducedPerSample(nn.Module):

    def __init__(self, loss, reduction):
        super().__init__()
        self.loss = loss
        self.reduction = reduction

    def forward(self, *args, **kwargs):
        loss_results = self.loss(*args, **kwargs)
        n_dims = len(loss_results.shape)
        if n_dims > 1:
            dims_to_reduce = tuple(range(1, n_dims))
            return self.reduction(loss_results, dim=dims_to_reduce, keepdim=True)
        return loss_results


class LossFlowData(IFeaturesDict):

    def __init__(self, outputs, targets):
        self.outputs = outputs
        self.targets = targets

    # To be compatible with the pipeline
    def __getattr__(self, item):
        return self


class LossItems(IOut, IFlowTaskResult, ICondition):

    def __init__(self, loss_items):
        self.loss_items = loss_items

    def __iadd__(self, other):
        return LossItems(self.loss_items + other.loss_items)

    # None of the methods below modify the state. They are here
    # to be compatible with the pipeline
    def __getattr__(self, item):
        return self

    def __or__(self, result):
        return self

    def __invert__(self):
        return self

    def __and__(self, other):
        return self


def get_flow_data(*args, **kwargs):
    if len(args) == 2:
        return LossFlowData(*args)

    for arg in args:
        if isinstance(arg, LossFlowData):
            return arg

    for arg in kwargs.values():
        if isinstance(arg, LossFlowData):
            return arg


def squeeze_if_needed(tensor):
    if len(tensor.shape) > 2:
        raise ValueError(f'Trying to squeeze the second dimension out of a tensor with shape: {tensor.shape}')
    if len(tensor.shape) == 2:
        return tensor[:, 0]
    return tensor


class BaseMetricDecorator(nn.Module):

    def __init__(self, task_name, available_func, prefix, metric):
        super().__init__()
        self.task_name = task_name
        self.available = available_func
        self.prefix = prefix
        self.metric = metric

    def forward(self, *args, **kwargs):
        loss_flow_data = get_flow_data(*args, **kwargs)
        return self.compute_with_precondition(loss_flow_data, self.metric)

    def compute_with_precondition(self, loss_flow_data, metric):
        key = self.prefix + self.task_name
        outputs = loss_flow_data.outputs[key]
        precondition = loss_flow_data.outputs[f'precondition|{key}']
        targets = loss_flow_data.targets[key]
        loss_items = torch.zeros(1, dtype=outputs.dtype, device=outputs.device)
        if precondition.sum() == 0:
            return loss_items
        precondition = squeeze_if_needed(precondition)
        metric_res = metric(outputs[precondition], targets[precondition])
        return self.postprocess_results(loss_items, metric_res, precondition)

    def postprocess_results(self, loss_items, metric_res, precondition):
        return metric_res


class TaskLossDecorator(BaseMetricDecorator):

    def __init__(self, task_name, available_func, prefix, loss):
        super().__init__(task_name, available_func, prefix, loss)

    def postprocess_results(self, loss_items, metric_res, precondition):
        return LossItems(metric_res)


class TaskFlowLoss(nn.Module):

    def __init__(self, task_flow, prefix=''):
        super().__init__()
        self._task_flow = task_flow
        # Save a reference to the flow function of the original class
        # We will then call it by replacing the self, this way effectively running
        # it with this class. And this class stores Pytorch modules as class attributes
        self.flow = task_flow.get_flow_func()

        for key, task in task_flow.tasks.items():
            if not task.has_children():
                instance = TaskLossDecorator(task.get_name(), task.get_available_func(), prefix, task.get_loss())
            else:
                instance = TaskFlowLoss(task, prefix=f'{prefix}{task.get_name()}.')

            setattr(self, key, instance)

    def forward(self, *args):
        """
        TaskFlowLoss can be invoked either by giving two arguments: (outputs, targets), or bby giving a single
        LossFlowData argument, which holds the outputs and the targets.
        :param args:
        :return:
        """
        is_root = len(args) == 2
        if is_root:
            outputs, targets = args
        else:
            loss_flow_data = args[0]
            outputs = loss_flow_data.outputs
            targets = loss_flow_data.targets

        value = any_value(outputs)
        loss_items = torch.zeros(1, dtype=value.dtype, device=value.device)
        flow_result = self.flow(self, LossFlowData(outputs, targets), LossItems(loss_items))

        if not is_root:
            return LossItems(flow_result.loss_items)

        return flow_result.loss_items

    def get_leaf_losses(self):
        all_losses = {}
        for key, task in self._task_flow.tasks.items():
            child_loss = getattr(self, key)
            if task.has_children():
                all_losses.update(child_loss.get_leaf_losses())
            else:
                path = child_loss.prefix + child_loss.task_name
                all_losses[path] = child_loss
        return all_losses

    def get_metrics(self):
        all_metrics = []
        for key, task in self._task_flow.tasks.items():
            child_loss = getattr(self, key)
            if task.has_children():
                all_metrics += child_loss.get_metrics()
            else:
                for metric_name, metric in task.get_metrics():
                    metric_decorator = BaseMetricDecorator(task.get_name(),
                                                           task.get_available_func(),
                                                           child_loss.prefix,
                                                           metric)
                    all_metrics.append((metric_name, metric_decorator))
        return all_metrics

    def catalyst_callbacks(self):
        from catalyst.core import MetricCallback
        callbacks = []
        for path, loss in self.get_leaf_losses().items():
            metric_decorator = BaseMetricDecorator(loss.task_name,
                                                   loss.available,
                                                   loss.prefix,
                                                   loss.metric)
            callbacks.append(MetricCallback(f'loss_{path}', metric_decorator))
        for metric_name, metric_decorator in self.get_metrics():
            full_name = f'{metric_name}_{metric_decorator.prefix}{metric_decorator.task_name}'
            callback = MetricCallback(full_name, metric_decorator)
            callbacks.append(callback)
        return callbacks


class TaskFlowLossPerSample(nn.Module):

    def __init__(self, task_flow, prefix=''):
        super().__init__()

        self._all_children = task_flow.get_all_children(prefix=prefix)
        self._all_losses = self._collect_leaf_losses_per_sample()

    def forward(self, outputs, targets):
        res = {}
        value = any_value(outputs)
        bs = len(value)
        overall_loss_items = torch.zeros(bs, device=value.device, dtype=value.dtype)
        for path, loss in self._all_losses.items():
            loss_items = loss(outputs, targets).loss_items
            res[path] = loss_items.squeeze(dim=-1)
            indices = torch.arange(bs, device=value.device)
            precondition = outputs[f'precondition|{path}']
            axes = tuple(range(1, len(precondition.shape)))
            if len(axes) > 0:
                precondition = precondition.sum(axis=axes) > 0
            res[f'indices|{path}'] = indices[precondition]
            overall_loss_items[precondition] += res[path]
        res['overall'] = overall_loss_items
        res['indices|overall'] = torch.arange(bs, device=value.device)
        return res

    def get_leaf_losses_per_sample(self):
        return self._all_losses

    def _collect_leaf_losses_per_sample(self):
        all_losses = {}
        for path, task in self._all_children.items():
            if '.' not in path:
                prefix = ''
            else:
                prefix = path.rsplit('.', 1)[0] + '.'
            all_losses[path] = TaskLossDecorator(task.get_name(),
                                                 task.get_available_func(),
                                                 prefix,
                                                 task.get_per_sample_loss())
        return all_losses



from functools import partial
from typing import Optional, Tuple, Callable, Union

from torch import nn

from dnn_cool.decoders import Decoder
from dnn_cool.metrics import TorchMetric


class To:
    """Syntactic sugar for task converters. You can achieve the same through `functools.partial`, but here you can
    give a supplier for some fields, so that the creation of the respective field is delayed."""

    def __init__(self,
                 task_cls,
                 module_supplier: Callable[[], nn.Module] = None,
                 loss_supplier: Callable[[], nn.Module] = None,
                 per_sample_loss_supplier: Callable[[], nn.Module] = None,
                 decoder_supplier: Callable[[], Decoder] = None,
                 available_func=None,
                 inputs=None,
                 activation: Optional[nn.Module] = None,
                 metrics: Tuple[str, TorchMetric] = None):
        self.task_cls = task_cls
        self.module_supplier = module_supplier
        self.loss_supplier = loss_supplier
        self.per_sample_loss_supplier = per_sample_loss_supplier
        self.decoder_supplier = decoder_supplier
        self.available_func = available_func
        self.inputs = inputs
        self.activation = activation
        self.metrics = metrics

    def __call__(self, name, labels):
        kwargs = {}
        if self.module_supplier is not None:
            kwargs['module'] = self.module_supplier()
        if self.loss_supplier is not None:
            kwargs['loss'] = self.loss_supplier()
        if self.per_sample_loss_supplier is not None:
            kwargs['per_sample_loss'] = self.per_sample_loss_supplier()
        if self.decoder_supplier is not None:
            kwargs['decoder'] = self.decoder_supplier()
        if self.available_func is not None:
            kwargs['available_func'] = self.available_func
        if self.inputs is not None:
            kwargs['inputs'] = self.inputs
        if self.activation is not None:
            kwargs['activation'] = self.activation

        return self.task_cls(name=name, labels=labels, **kwargs)


class EfficientNetHead:
    in_features_dict = {
        "efficientnet-b0": 1280,
        "efficientnet-b1": 1280,
        "efficientnet-b2": 1408,
        "efficientnet-b3": 1536,
        "efficientnet-b4": 1792,
        "efficientnet-b5": 2048,
        "efficientnet-b6": 2304,
        "efficientnet-b7": 2560,
        "efficientnet-b8": 2816
    }

    @staticmethod
    def create(model: Union[str, int], out_features, bias=True):
        return EfficientNetHead.create_supplier(model, out_features, bias)()

    @staticmethod
    def create_supplier(model: Union[str, int], out_features, bias=True):
        if isinstance(model, int):
            model = f'efficientnet-b{model}'
        in_features = EfficientNetHead.in_features_dict[model]
        return partial(nn.Linear, in_features=in_features, out_features=out_features, bias=bias)


class ToEfficient(To):

    def __init__(self,
                 task_cls,
                 model: Union[str, int],
                 n_classes: int,
                 bias=True,
                 loss_supplier: Callable[[], nn.Module] = None,
                 per_sample_loss_supplier: Callable[[], nn.Module] = None,
                 decoder_supplier: Callable[[], Decoder] = None,
                 available_func=None,
                 inputs=None,
                 activation: Optional[nn.Module] = None,
                 metrics: Tuple[str, TorchMetric] = None):
        module_supplier = EfficientNetHead.create_supplier(model, n_classes, bias)
        super().__init__(task_cls,
                         module_supplier,
                         loss_supplier,
                         per_sample_loss_supplier,
                         decoder_supplier,
                         available_func,
                         inputs,
                         activation,
                         metrics)



from dataclasses import dataclass
from treelib import Tree

from dnn_cool.utils import any_value


def find_results_for_treelib(*args, **kwargs):
    for arg in args:
        if isinstance(arg, Results):
            return arg

    for arg in kwargs.values():
        if isinstance(arg, Results):
            return arg


@dataclass
class Precondition:
    path: str

    def __invert__(self):
        return self


class TreeExplanation:

    def __init__(self, tree, start_node, results, prefix):
        self.tree = tree
        self.start_node = start_node
        self.results = results
        self.precondition = None
        self.prefix = prefix

    def __iadd__(self, other):
        if self.start_node is None:
            return other
        parent = self.start_node.identifier if other.precondition is None else other.precondition.path
        if parent in self.tree:
            self.tree.paste(parent, other.tree)
        return self

    def __getattr__(self, item):
        return Precondition(self.prefix + item)

    def __or__(self, precondition: Precondition):
        self.precondition = precondition
        return self


class LeafExplainer:

    def __init__(self, task, prefix):
        self.prefix = prefix
        self.task_name = task.get_name()

    def __call__(self, *args, **kwargs):
        results: Results = find_results_for_treelib(*args, **kwargs)
        path = self.prefix + self.task_name

        decoded = results.module_output.decoded[path][results.idx].detach().cpu().numpy()
        activated = results.module_output.activated[path][results.idx].detach().cpu().numpy()
        logits = results.module_output.logits[path][results.idx].detach().cpu().numpy()
        precondition = results.module_output.preconditions.get(path)
        if precondition is not None:
            precondition = precondition[results.idx][0].item()

        tree = Tree()
        should_create_node = (precondition is None) or (precondition is True)
        if should_create_node:
            description = f'{self.task_name} | decoded: {decoded}, activated: {activated}, logits: {logits}'
            start_node = tree.create_node(description, f'inp_{results.idx}.{path}')
        else:
            start_node = None
        return TreeExplanation(tree, start_node, results, self.prefix)


class Results:

    def __init__(self, module_output, idx=None):
        self.module_output = module_output
        self.idx = idx

    # Pipeline compatibility
    def __getattr__(self, item):
        return self


class TreeExplainer:

    def __init__(self, task_flow, prefix=''):
        self.task_flow = task_flow

        self.flow = task_flow.get_flow_func()
        self.prefix = prefix
        self.task_name = task_flow.get_name()

        for key, task in task_flow.tasks.items():
            if not task.has_children():
                instance = LeafExplainer(task, prefix)
            else:
                instance = TreeExplainer(task, prefix=f'{prefix}{task.get_name()}.')
            setattr(self, key, instance)

    def __call__(self, x):
        if not isinstance(x, Results):
            x = Results(x)

        if x.idx is None:
            n = len(any_value(x.module_output.logits))
            tree = Tree()
            batch_node = tree.create_node(tag='batch', identifier='batch')
            for i in range(n):
                inp_tree = Tree()
                inp_node = inp_tree.create_node(tag=f'inp {i}', identifier=f'inp_{i}.{self.prefix}')
                x_for_id = Results(x.module_output, i)
                out = TreeExplanation(inp_tree, inp_node, x_for_id, f'inp_{i}.')
                out = self.flow(self, x_for_id, out)
                tree.paste(batch_node.identifier, out.tree)

            if len(self.prefix) == 0:
                return tree
            return TreeExplanation(tree, start_node=batch_node, results=x, prefix=self.prefix)

        tree = Tree()
        start_node = tree.create_node(tag=self.task_name, identifier=f'inp_{x.idx}.{self.prefix}.{self.task_name}')
        out = TreeExplanation(tree, start_node=start_node, results=x, prefix=f'inp_{x.idx}.{self.prefix}')
        out = self.flow(self, x, out)

        no_new_nodes_added = (len(out.tree.nodes) == 1) and (out.start_node.identifier in out.tree)
        # if all the nodes of a nested flow are empty, remove the whole flow node.
        if no_new_nodes_added:
            out.tree = Tree()
        if len(self.prefix) == 0:
            return out.tree
        return out



from dataclasses import dataclass, field
from pathlib import Path
from typing import Callable, Dict, Optional, Tuple, List

import numpy as np
from catalyst.contrib.tools.tensorboard import SummaryWriter
from catalyst.core import Callback, CallbackOrder, State
from torch.utils.data import Dataset, SequentialSampler

from dnn_cool.task_flow import TaskFlow
from dnn_cool.utils import any_value


def publish_all(prefix, sample, mapping_key, key, writer, mapping, task_name):
    if mapping_key in mapping:
        publishers = mapping[mapping_key]
        for publisher in publishers:
            publisher(writer, sample, prefix, task_name, key)


def img(writer: SummaryWriter, sample: Tuple, prefix: str, task_name: str, key: str):
    X, y = sample
    writer.add_image(f'{prefix}_{task_name}_images', X[key])


def text(writer: SummaryWriter, sample: Tuple, prefix: str, task_name: str, key: str):
    X, y = sample
    writer.add_text(f'{prefix}_{task_name}_text', X[key])


def default_tensorboard_type_mapping():
    return {
        'img': [img],
        'text': [text]
    }


@dataclass()
class TensorboardConverter:
    col_mapping: Dict[str, List[Callable]] = field(default_factory=lambda: {})
    type_mapping: Dict[str, List[Callable]] = field(default_factory=default_tensorboard_type_mapping)
    col_to_type_mapping: Dict[str, str] = field(default_factory=lambda: {})

    def __call__(self, writer: SummaryWriter, sample: Tuple, prefix: str, task_name: str):
        if task_name == 'gt':
            return
        X, y = sample
        for key in X:
            publish_all(prefix, sample, key, key, writer, self.col_mapping, task_name)
        for key in X:
            if key in self.col_to_type_mapping:
                publish_all(prefix, sample, self.col_to_type_mapping[key], key, writer, self.type_mapping, task_name)


@dataclass
class TensorboardConverters:
    logdir: Path
    datasets: Dict[str, Dataset]
    tensorboard_loggers: Callable = field(default_factory=TensorboardConverter)
    loggers: Dict[str, SummaryWriter] = field(default_factory=lambda: {})
    top_k: int = 10

    def initialize(self, state):
        if (self.logdir is not None) and (state.loader_name not in self.loggers):
            path = str(self.logdir / f"{state.loader_name}_log")
            writer = SummaryWriter(path)
            self.loggers[state.loader_name] = writer

    def publish(self, state, interpretations):
        for key, value in interpretations.items():
            if key.startswith('indices'):
                continue
            sorted_indices = value.argsort()
            best_indices = interpretations[f'indices|{key}'][sorted_indices][:self.top_k]
            worst_indices = interpretations[f'indices|{key}'][sorted_indices][-self.top_k:]
            writer: SummaryWriter = self.loggers[state.loader_name]
            dataset = self.datasets[state.loader_name]
            self._publish_inputs(best_indices, writer, dataset, prefix='best', key=key)
            self._publish_inputs(worst_indices, writer, dataset, prefix='worst', key=key)

    def _publish_inputs(self, best_indices, writer, dataset, prefix, key):
        for idx in best_indices:
            if self.tensorboard_loggers is not None:
                self.tensorboard_loggers(writer, dataset[idx], prefix, key)

    def close(self, state):
        """Close opened tensorboard writers"""
        if state.logdir is None:
            return

        for logger in self.loggers.values():
            logger.close()


class InterpretationCallback(Callback):
    def __init__(self, flow: TaskFlow, tensorboard_converters: Optional[TensorboardConverters] = None):
        super().__init__(CallbackOrder.Metric)
        self.flow = flow

        self.overall_loss = flow.get_per_sample_loss()
        self.leaf_losses = self.overall_loss.get_leaf_losses_per_sample()
        self.interpretations = {}
        self.loader_counts = {}

        self.tensorboard_converters = tensorboard_converters

    def _initialize_interpretations(self):
        interpretation_dict = {
            'overall': [],
            'indices|overall': []
        }
        for path in self.leaf_losses:
            interpretation_dict[path] = []
            interpretation_dict[f'indices|{path}'] = []
        return interpretation_dict

    def on_loader_start(self, state: State):
        if not isinstance(state.loaders[state.loader_name].sampler, SequentialSampler):
            return
        self.interpretations[state.loader_name] = self._initialize_interpretations()
        self.loader_counts[state.loader_name] = 0

        if self.tensorboard_converters is not None:
            self.tensorboard_converters.initialize(state)

    def on_batch_end(self, state: State):
        if not isinstance(state.loaders[state.loader_name].sampler, SequentialSampler):
            return
        outputs = state.output['logits']
        targets = state.input['targets']
        overall_res = self.overall_loss(outputs, targets)
        start = self.loader_counts[state.loader_name]
        bs = len(any_value(outputs))

        for path, loss in overall_res.items():
            if path.startswith('indices'):
                continue
            self.interpretations[state.loader_name][path].append(loss.detach().cpu().numpy())
            ind_key = f'indices|{path}'
            indices = overall_res[ind_key] + start
            self.interpretations[state.loader_name][ind_key].append(indices.detach().cpu().numpy())
        self.loader_counts[state.loader_name] += bs

    def on_loader_end(self, state: State):
        if not isinstance(state.loaders[state.loader_name].sampler, SequentialSampler):
            return
        self.interpretations[state.loader_name] = {
            key: np.concatenate(value, axis=0)
            for key, value in self.interpretations[state.loader_name].items()
        }

        if self.tensorboard_converters is not None:
            self.tensorboard_converters.publish(state, self.interpretations[state.loader_name])

    def on_stage_end(self, state: State):
        if not isinstance(state.loaders[state.loader_name].sampler, SequentialSampler):
            return
        if self.tensorboard_converters is not None:
            self.tensorboard_converters.close(state)
